/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: ip6.c 4111 2014-02-10 14:09:31Z masar@masar $
#################################################################################################*/

static char Psz_SvnId[] __attribute__ ((unused)) =
    "$Id: ip6.c 4111 2014-02-10 14:09:31Z masar@masar $";

#include "ip6.h"
#include "network.h"
#include "util.h"

#include <errno.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <string.h>
#include <syslog.h>

//************************************************************************************************/
//* Data
//************************************************************************************************/
static void ip6SendIcmpEcho (PatdClient * pd);

/*************************************************************************************************/
/** Writes IP6 response.
**************************************************************************************************/
static int ip6SendClientResponse (PatdClient * pd, uint8_t iResult,
                                  uint8_t * pAddr, uint8_t iCurSeq, uint32_t lMSec)
   {
   int iCount = 0;
   aui8_Buffer[iCount++] = iResult;
   aui8_Buffer[iCount++] = (iCurSeq     >>  0) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >> 24) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >> 16) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >>  8) & 0xFF;
   aui8_Buffer[iCount++] = (lMSec       >>  0) & 0xFF;

   for (int j = 0; j < 16; j += 1) {
      aui8_Buffer[iCount++] = pAddr[j];
      }

   if (iCount != RESPONSE_IP6_SIZE) {
      syslog(LOG_ERR, "ip6SendClientResponse: invalid size");
      deletePatdClient(pd);
      return 1;
      }

   if (writeAll(pd->fdClient, aui8_Buffer, RESPONSE_IP6_SIZE)) {
      syslog(LOG_WARNING, "ip6SendClientResponse: cannot write to client: socket=%i id=%i",
                          pd->fdClient, pd->iId);
      deletePatdClient(pd);
      return 1;
      }
   syslog(LOG_DEBUG, "ip6SendClientResponse: ok");
   return 0;
   }

/*************************************************************************************************/
/** Writes ip6 response.
**************************************************************************************************/
static void ip6WriteClientError (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip6WriteClientError: socket=%i id=%i status=%i",
                      pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP6) {
      syslog(LOG_ERR, "ip6WriteClientError: invalid status");
      return;
      }

   struct timespec tsNow;
   struct timespec tsTime;

   clock_gettime(CLOCK_MONOTONIC, &tsNow);
   tsDelta(&tsNow, &pd->tsLast, &tsTime);

   uint8_t iResult = ((pd->iCurRet < pd->iMaxRet) && (pd->iCurSeq < pd->iMaxSeq)) ?
                        RESULT_ERROR :
                        RESULT_ERROR_LAST;
   uint32_t lMSec = tsTime.tv_sec * 1000000000L + tsTime.tv_nsec;
   uint8_t * pAddr = (uint8_t *) ((struct sockaddr_in6 *) &pd->saAddr)->sin6_addr.s6_addr;

   if (ip6SendClientResponse(pd, iResult, pAddr, pd->iCurSeq, lMSec)) {
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   if (iResult == RESULT_ERROR) {
      ip6SendIcmpEcho(pd);
      }
   else {
      pd->iStatus = STATUS_NEW;
      rearmClient(pd);
      }

   syslog(LOG_DEBUG, "ip6WriteClientError: ok");
   }

/*************************************************************************************************/
/** Sends an ICMP4 ECHO.
**************************************************************************************************/
void ip6SendIcmpEcho (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip6SendIcmpEcho: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP6) {
      syslog(LOG_DEBUG, "ip6SendIcmpEcho: Invalid status");
      deletePatdClient(pd);
      return;
      }

   if ((pd->iCurRet >= pd->iMaxRet) || (pd->iCurSeq >= pd->iMaxSeq)) {
      syslog(LOG_DEBUG, "ip6SendIcmpEcho: no more");
      pd->iStatus = STATUS_NEW;                          // ???
      rearmClient(pd);
      return;
      }

   pd->iCurSeq += 1;
   pd->iCurTTL = (pd->iCurTTL >= pd->iMaxTTL) ? pd->iCurTTL : pd->iCurTTL + 1;

   ssize_t iSize = pd->iSize + sizeof(struct icmp6_hdr);
   struct icmp6_hdr * pIcmp6 = (struct icmp6_hdr *) aui8_Buffer;

   memset(aui8_Buffer, 0, iSize);
   aui8_Buffer[iSize-1] = 0xFF;

   pIcmp6->icmp6_type = ICMP6_ECHO_REQUEST;
   pIcmp6->icmp6_code = 0;
   pIcmp6->icmp6_cksum = 0;
   pIcmp6->icmp6_id = ui32_Pid;
   pIcmp6->icmp6_seq = (pd->iId << MAX_SEQ_BITS) | pd->iCurSeq;

   int iVal = pd->iFragment ? IPV6_PMTUDISC_DONT : IPV6_PMTUDISC_DO;
   if (setsockopt(patd_Icmp6.fdIcmp, IPPROTO_IPV6, IPV6_MTU_DISCOVER, &iVal, sizeof(iVal)) < 0) {
      syslog(LOG_WARNING, "ip6SendIcmpEcho: set Socket Frag error %i %s", pd->fdClient, strerror(errno));
      deletePatdClient(pd);
      return;
      }

   if (setsockopt(patd_Icmp6.fdIcmp, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &pd->iCurTTL, sizeof(pd->iCurTTL)) < 0) {
      syslog(LOG_WARNING, "ip6SendIcmpEcho: set Socket TTL error %i %s", pd->fdClient, strerror(errno));
      deletePatdClient(pd);
      return;
      }

   if (sendto(patd_Icmp6.fdIcmp, aui8_Buffer, iSize, 0,
              (struct sockaddr *) &pd->saAddr, sizeof(pd->saAddr)) != iSize) {
      syslog(LOG_DEBUG, "ip6SendIcmpEcho: write IP6 error %i %s", pd->fdClient, strerror(errno));
      pd->iCurSeq = pd->iMaxSeq;       // Force Last
      ip6WriteClientError(pd);
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   syslog(LOG_DEBUG, "ip6SendIcmpEcho: done echo.id=%x echo.seq=%x",
                     pIcmp6->icmp6_id, pIcmp6->icmp6_seq);
   }

/*************************************************************************************************/
/** Reads a Client IP6 Config.
**************************************************************************************************/
void ip6ReadConfig (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip6ReadConfig: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_READ_IP6) {
      syslog(LOG_ERR, "ip6ReadConfig: not STATUS_READ_IP6");
      return;
      }

   ssize_t iSize = recv(pd->fdClient, aui8_Buffer, sizeof(aui8_Buffer), MSG_PEEK);
   if (iSize < COMMAND_IP6_SIZE-1) {
      syslog(LOG_DEBUG, "ip6ReadConfig: no enought data size=%zi", iSize);
      return;
      }

   if (readAll(pd->fdClient, aui8_Buffer, COMMAND_IP6_SIZE-1)) {
      syslog(LOG_WARNING, "ip6ReadConfig: error read socket=%i id=%i", pd->fdClient, pd->iId);
      deletePatdClient(pd);
      return;
      }

   int iCount = 0;
   pd->iStatus = STATUS_IP6;

   pd->iMaxSeq = aui8_Buffer[iCount++];
   pd->iMaxSeq = (pd->iMaxSeq < 1) || (pd->iMaxSeq > MAX_SEQ) ?  3 : pd->iMaxSeq;
   pd->iCurSeq = 0;

   pd->iMaxRet = aui8_Buffer[iCount++];
   pd->iMaxRet = (pd->iMaxRet < 1) || (pd->iMaxRet > MAX_SEQ) ?  1 : pd->iMaxRet;
   pd->iCurRet = 0;

   pd->iCurTTL = aui8_Buffer[iCount++];
   pd->iCurTTL = ((pd->iCurTTL < 0) || (pd->iCurTTL > MAX_TTL)) ? MAX_TTL : pd->iCurTTL;

   pd->iMaxTTL = aui8_Buffer[iCount++];
   pd->iMaxTTL = ((pd->iMaxTTL < 1) || (pd->iMaxTTL > MAX_TTL)) ? MAX_TTL : pd->iMaxTTL;

   pd->iCurTTL = (pd->iCurTTL > pd->iMaxTTL) ? pd->iMaxTTL : pd->iCurTTL;

   pd->iSize   = (aui8_Buffer[iCount + 0] <<  8) +
                 (aui8_Buffer[iCount + 1] <<  0);
   iCount += 2;
   pd->iSize   = ((pd->iSize < 0) || (pd->iSize > MAX_ICMP_SIZE)) ? 0 : pd->iSize;

   pd->iFragment = aui8_Buffer[iCount++] ? 1 : 0;

   int iTimeOut = (aui8_Buffer[iCount + 0] <<  8) +
                  (aui8_Buffer[iCount + 1] <<  0);
   iCount += 2;
   if ((iTimeOut <= 0) || (iTimeOut > 2000)) {
      iTimeOut = 1000;
      }

   pd->tsTimeOut.tv_sec  = iTimeOut / 1000;
   pd->tsTimeOut.tv_nsec = (iTimeOut % 1000) * 1000000L;

   struct sockaddr_in6 * saAddr6 = (struct sockaddr_in6 *) &pd->saAddr;
   saAddr6->sin6_family = AF_INET6;
   saAddr6->sin6_port = 0;
   saAddr6->sin6_addr.s6_addr[ 0] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 1] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 2] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 3] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 4] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 5] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 6] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 7] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 8] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[ 9] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[10] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[11] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[12] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[13] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[14] = aui8_Buffer[iCount++];
   saAddr6->sin6_addr.s6_addr[15] = aui8_Buffer[iCount++];

   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   syslog(LOG_DEBUG, "ip6ReadConfig: socket=%i id=%i status=%i"
                     " ip=%02X%02X:%02X%02X:%02X%02X:%02X%02X:"
                         "%02X%02X:%02X%02X:%02X%02X:%02X%02X"
                     " seq=%i ret=%i curttl=%i maxttl=%i"
                     " size=%i frag=%i to=%li.%09li",
                     pd->fdClient, pd->iId, pd->iStatus,
                     saAddr6->sin6_addr.s6_addr[ 0], saAddr6->sin6_addr.s6_addr[ 1],
                     saAddr6->sin6_addr.s6_addr[ 2], saAddr6->sin6_addr.s6_addr[ 3],
                     saAddr6->sin6_addr.s6_addr[ 4], saAddr6->sin6_addr.s6_addr[ 5],
                     saAddr6->sin6_addr.s6_addr[ 6], saAddr6->sin6_addr.s6_addr[ 7],
                     saAddr6->sin6_addr.s6_addr[ 8], saAddr6->sin6_addr.s6_addr[ 9],
                     saAddr6->sin6_addr.s6_addr[10], saAddr6->sin6_addr.s6_addr[11],
                     saAddr6->sin6_addr.s6_addr[12], saAddr6->sin6_addr.s6_addr[13],
                     saAddr6->sin6_addr.s6_addr[14], saAddr6->sin6_addr.s6_addr[15],
                     pd->iMaxSeq, pd->iMaxRet,
                     pd->iCurTTL, pd->iMaxTTL, pd->iSize, pd->iFragment,
                     pd->tsTimeOut.tv_sec, pd->tsTimeOut.tv_nsec);

   ip6SendIcmpEcho(pd);
   }

/*************************************************************************************************/
/** Writes ip6 response.
**************************************************************************************************/
static void ip6WriteClientOk (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip6WriteClientOk: socket=%i id=%i status=%i",
                      pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP6) {
      syslog(LOG_DEBUG, "ip6WriteClientOk: invalid status");
      return;
      }

   if (pd->iCurRet >= pd->iMaxRet) {
      syslog(LOG_DEBUG, "ip6WriteClientOk: yet sent MaxRet");
      return;
      }
   pd->iCurRet += 1;

   struct timespec tsNow;
   struct timespec tsTime;

   clock_gettime(CLOCK_MONOTONIC, &tsNow);
   tsDelta(&tsNow, &pd->tsLast, &tsTime);

   uint8_t iResult =((pd->iCurRet < pd->iMaxRet) && (pd->iCurSeq < pd->iMaxSeq)) ?
                        RESULT_OK :
                        RESULT_OK_LAST;
   uint32_t lMSec = tsTime.tv_sec * 1000000000L + tsTime.tv_nsec;
   uint8_t * pAddr = (uint8_t *) ((struct sockaddr_in6 *) &pd->saAddr)->sin6_addr.s6_addr;

   if (ip6SendClientResponse(pd, iResult, pAddr, pd->iCurSeq, lMSec)) {
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   if (iResult == RESULT_OK) {
      ip6SendIcmpEcho(pd);
      }
   else {
      pd->iStatus = STATUS_NEW;
      rearmClient(pd);
      }

   syslog(LOG_DEBUG, "ip6WriteClientOk: ok");
   }

/*************************************************************************************************/
/** Writes ip6 response.
**************************************************************************************************/
static void ip6WriteClientTrace (PatdClient * pd, uint8_t * pAddr)
   {
   syslog(LOG_DEBUG, "ip6WriteClientTrace: socket=%i id=%i status=%i",
                      pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_IP6) {
      syslog(LOG_ERR, "ip6WriteClientTrace: invalid status");
      return;
      }

   struct timespec tsNow;
   struct timespec tsTime;

   clock_gettime(CLOCK_MONOTONIC, &tsNow);
   tsDelta(&tsNow, &pd->tsLast, &tsTime);

   uint8_t iResult = ((pd->iCurRet < pd->iMaxRet) && (pd->iCurSeq < pd->iMaxSeq)) ?
                        RESULT_TRACE :
                        RESULT_TRACE_LAST;
   uint32_t lMSec = tsTime.tv_sec * 1000000000L + tsTime.tv_nsec;

   if (ip6SendClientResponse(pd, iResult, pAddr, pd->iCurSeq, lMSec)) {
      return;
      }
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);

   if (iResult == RESULT_TRACE) {
      ip6SendIcmpEcho(pd);
      }
   else {
      pd->iStatus = STATUS_NEW;
      rearmClient(pd);
      }

   syslog(LOG_DEBUG, "ip6WriteClientTrace: ok");
   }

/*************************************************************************************************/
/** ip6TimeExceeded.
**************************************************************************************************/
static void ip6TimeExceeded (struct sockaddr_in6 * pAddr, struct icmp6_hdr * pIcmpOrg)
   {
   syslog(LOG_DEBUG, "ip6TimeExceeded: start");

   int iId = pIcmpOrg->icmp6_seq >> MAX_SEQ_BITS;
   int iCurSeq = pIcmpOrg->icmp6_seq & MAX_SEQ_MASK;

   if (iId >= MAX_CLIENTS) {
      syslog(LOG_WARNING, "ip6TimeExceeded: invalid id: %i", iId);
      return;
      }
   syslog(LOG_DEBUG, "ip6TimeExceeded: id=%i", iId);

   PatdClient * pd = getPatdClient(iId);
   if (pd->iStatus != STATUS_IP6) {
      syslog(LOG_WARNING, "ip6TimeExceeded: strange status socket=%i id=%i status=%i",
                        pd->fdClient, pd->iId, pd->iStatus);
      deletePatdClient(pd);
      return;
      }

   if (pd->iCurSeq != iCurSeq) {
      syslog(LOG_DEBUG, "ip6TimeExceeded: old seq: socket=%i id=%i seq %i!=%i",
                        pd->fdClient, pd->iId, pd->iCurSeq, iCurSeq);
      return;
      }
   ip6WriteClientTrace(pd, pAddr->sin6_addr.s6_addr);

   syslog(LOG_DEBUG, "ip6TimeExceeded: ok");
   }

/*************************************************************************************************/
/** ip6EchoReply.
**************************************************************************************************/
static void ip6EchoReply (struct icmp6_hdr * pIcmp)
   {
   syslog(LOG_DEBUG, "ip6EchoReply: start");
//   syslogDump(LOG_DEBUG, "ECHO4 REPLY DUMP: ", pIcmp, iSize);
//   syslog(LOG_DEBUG, "ECHO4 REPLY: %x %x", pIcmp->un.echo.id, pIcmp->un.echo.sequence);

   int iId = pIcmp->icmp6_seq >> MAX_SEQ_BITS;
   int iCurSeq = pIcmp->icmp6_seq & MAX_SEQ_MASK;

   if (iId >= MAX_CLIENTS) {
      syslog(LOG_WARNING, "ip6EchoReply: invalid id: %i", iId);
      return;
      }
   syslog(LOG_DEBUG, "ip6EchoReply: id=%i", iId);

   PatdClient * pd = getPatdClient(iId);
   if (pd->iStatus != STATUS_IP6) {
      syslog(LOG_DEBUG, "ip6EchoReply: strange status socket=%i id=%i status=%i",
                        pd->fdClient, pd->iId, pd->iStatus);
      deletePatdClient(pd);
      return;
      }

   if (pd->iCurSeq != iCurSeq) {
      syslog(LOG_DEBUG, "ip6EchoReply: old seq: socket=%i id=%i seq %i!=%i",
                        pd->fdClient, pd->iId, pd->iCurSeq, iCurSeq);
      return;
      }
   ip6WriteClientOk(pd);

   syslog(LOG_DEBUG, "ip6EchoReply: ok");
   }

/*************************************************************************************************/
/** ip6Process.
**************************************************************************************************/
void ip6Process (void)
   {
//   syslog(LOG_DEBUG, "processIcmp6: start");
//   ssize_t iSize = recv(patd_Icmp6.fdIcmp, aui8_Buffer, sizeof(aui8_Buffer), 0);
   struct sockaddr_storage saSrc;
   struct iovec iov[1];
   struct msghdr mhMessage;

   iov[0].iov_base = aui8_Buffer;
   iov[0].iov_len  = sizeof(aui8_Buffer);
   mhMessage.msg_name = &saSrc;
   mhMessage.msg_namelen = sizeof(saSrc);
   mhMessage.msg_iov = iov;
   mhMessage.msg_iovlen = 1;
   mhMessage.msg_control = 0;
   mhMessage.msg_controllen = 0;

   ssize_t iSize = recvmsg(patd_Icmp6.fdIcmp, &mhMessage, 0);
   while (iSize > 0) {
      syslog(LOG_DEBUG, "processIcmp6: read %zi", iSize);
//      syslogDump(LOG_DEBUG, "processIcmp6: ", aui8_Buffer, iSize);
//      syslogDump(LOG_DEBUG, "processIcmp6: ADDR ", &saSrc, mhMessage.msg_namelen);

      struct icmp6_hdr * pIcmp = (struct icmp6_hdr *)aui8_Buffer;
      struct ip6_hdr *pIpOrg = (struct ip6_hdr *)(pIcmp + 1);
      struct icmp6_hdr * pIcmpOrg = (struct icmp6_hdr *)(pIpOrg + 1); // TODO

      if ((pIcmp->icmp6_type == ICMP6_TIME_EXCEEDED) && (pIcmpOrg->icmp6_id == ui32_Pid)) {
         syslog(LOG_DEBUG, "processIcmp6: icmp ICMP6_TIME_EXCEEDED");
         ip6TimeExceeded((struct sockaddr_in6 *) &saSrc, pIcmpOrg);
         }
      else if ((pIcmp->icmp6_type == ICMP6_DST_UNREACH) && (pIcmpOrg->icmp6_id == ui32_Pid)) {
         syslog(LOG_DEBUG, "processIcmp6: icmp ICMP6_DST_UNREACH");
         ip6TimeExceeded((struct sockaddr_in6 *) &saSrc, pIcmpOrg);
         }
      else if ((pIcmp->icmp6_type == ICMP6_ECHO_REPLY) && (pIcmp->icmp6_id == ui32_Pid)) {
         syslog(LOG_DEBUG, "processIcmp6: icmp ICMP6_ECHO_REPLY");
         ip6EchoReply(pIcmp);
         }
      else {
         syslog(LOG_DEBUG, "processIcmp6: ignore type=%i code=%i id=%i size=%zi",
                           pIcmp->icmp6_type, pIcmp->icmp6_code, pIcmp->icmp6_id, iSize);
         }

      mhMessage.msg_namelen = sizeof(saSrc);
      iSize = recvmsg(patd_Icmp6.fdIcmp, &mhMessage, 0);
      }
//   syslog(LOG_DEBUG, "processIcmp6: done");
   }

/*************************************************************************************************/
/** Processed an ip6 TimeOut.
**************************************************************************************************/
void ip6TimeOut (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "ip6TimeOut: socket=%i id=%i status=%i", pd->fdClient, pd->iId, pd->iStatus);
   ip6WriteClientError(pd);
   }

/*************************************************************************************************/
/*************************************************************************************************/
