/*#################################################################################################
##  |\   ======   /|  ##     Copyright (C) 2013-2014  Maurizio Sartori     ##  |\   ======   /|  ##
##  |__\========/__|  ##  ===============================================  ##  |__\========/__|  ##
##  |\            /|  ##     Written by:  Maurizio Sartori                 ##  |\            /|  ##
##  |  \        /  |  ##     e-mail:      masar@MasarLabs.com              ##  |  \        /  |  ##
##  |    \    /    |  ##     Site:        http://www.masarlabs.com         ##  |    \    /    |  ##
##  |    /=\/=\    |  ##  ===============================================  ##  |    /=\/=\    |  ##
##  |  /========\  |  ##     License GPLv3+: GNU GPL version 3 or later    ##  |  /========\  |  ##
##  |/   ======   \|  ##          http://gnu.org/licenses/gpl.html         ##  |/   ======   \|  ##
###################################################################################################
## $Id: patd.c 4161 2014-02-18 16:31:21Z masar@masar $
#################################################################################################*/

static char Psz_SvnId[] __attribute__ ((unused)) =
    "$Id: patd.c 4161 2014-02-18 16:31:21Z masar@masar $";

#include "config.h"
#include "cap.h"
#include "patd.h"
#include "ip4.h"
#include "ip6.h"
#include "network.h"
#include "util.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/epoll.h>
#include <syslog.h>

//************************************************************************************************/
//* Data
//************************************************************************************************/
#define POOL_EVENTS 64

uint8_t    aui8_Buffer[IP_MAXPACKET + 1];
uint32_t   ui32_Pid;

PatdIcmp   patd_Icmp4;
PatdIcmp   patd_Icmp6;
PatdServer patd_Server;
PatdPool   patd_Pool;

static PatdClient apatd_Clients[MAX_CLIENTS];
static int        ai_ClientList[MAX_CLIENTS];
static int        i_ClientCount;
static struct timespec ts_ClientTimeOut;
struct epoll_event aep_Events[POOL_EVENTS];

static const char * str_Host = "127.0.0.1";
static const char * str_Port = "7865";
static const char * str_Pid = NULL;
static int i_LogLevel = LOG_NOTICE;
static int i_Foreground = 0;

/*************************************************************************************************/
/** Prototipes.
**************************************************************************************************/

/*************************************************************************************************/
/** Syslog Hexdump.
**************************************************************************************************/
//static void syslogDump (int iPriority, const char *strMsg, const void * pData, ssize_t iLen)
//   {
//   ssize_t iOffset = 0;
//   const char const * pStr = hexDump(pData, iLen, &iOffset);
//   while (pStr != NULL) {
//      syslog(iPriority, "%s%s", strMsg, pStr);
//      pStr = hexDump(pData, iLen, &iOffset);
//      }
//   }

/*************************************************************************************************/
/** Exit on error.
**************************************************************************************************/
static void chk (int iRc, const char * str)
   {
   if (iRc >= 0) {
      return;
      }
   syslog(LOG_ERR, "%s: %s", str, strerror(errno));
   syslog(LOG_NOTICE, "Program exit: %i", EXIT_FAILURE);
   exit(EXIT_FAILURE);
   }

/*************************************************************************************************/
/** Initializes.
**************************************************************************************************/
static void initializeData (void)
   {
   ui32_Pid = getpid();
   i_ClientCount = 0;

   ts_ClientTimeOut.tv_sec  = 5;
   ts_ClientTimeOut.tv_nsec = 0;

   for (int i = 0; i < MAX_CLIENTS; i += 1) {
      ai_ClientList[i] = i;

      apatd_Clients[i].fdClient = -1;
      apatd_Clients[i].iId = i;
      apatd_Clients[i].iStatus = STATUS_UNUSED;
      }
   }

/*************************************************************************************************/
/** Setup patd_Pool.
**************************************************************************************************/
static void setupPatdPool (void)
   {
   int fdPool = epoll_create1(0);
   chk(fdPool, "Failed to create epoll");

   patd_Pool.fdPool = fdPool;
   }

/*************************************************************************************************/
/** returns a PatdClient from its id.
**************************************************************************************************/
PatdClient * getPatdClient (int iId)
   {
   return &apatd_Clients[iId];
   }

/*************************************************************************************************/
/** Rearm a Client.
**************************************************************************************************/
void rearmClient (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "rearmClient: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);

   struct epoll_event epEvent;
   epEvent.data.ptr = pd;
   epEvent.events = EPOLLIN | EPOLLET;

   if (epoll_ctl(patd_Pool.fdPool, EPOLL_CTL_MOD, pd->fdClient, &epEvent)) {
      syslog(LOG_ERR, "rearmClient: cannot rearm epool: socket=%i id=%i",
                      pd->fdClient, pd->iId);
      deletePatdClient(pd);
      }
   }

/*************************************************************************************************/
/** Setup patd_Server.
**************************************************************************************************/
static void setupPatdServer (const char * strHost, const char * strPort)
   {
   int fdServer = openTcpServerSocket(strHost, strPort);
   chk(fdServer, "Failed to get server socket");
   chk(setSocketNonBlocking(fdServer), "Failed to set non blocking on server");

   patd_Server.fdServer = fdServer;

   struct epoll_event epEvent;
   epEvent.data.ptr = &patd_Server;
   epEvent.events = EPOLLIN | EPOLLET;

   chk(epoll_ctl(patd_Pool.fdPool, EPOLL_CTL_ADD, fdServer, &epEvent),
       "Failed to add fdServer to EPoll");
   }

/*************************************************************************************************/
/** Setup patd_Icmp.
**************************************************************************************************/
static void setupPatdIcmp (void)
   {
   int iVal = 1;
   struct epoll_event epEvent;

   chk(getCapabilities(), "Failed to get Capabilities");
   int fdIcmp4 = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
   int fdIcmp6 = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
   chk(dropCapabilities(), "Failed to drop Capabilities");

   chk(fdIcmp4, "Failed to get a raw IPv4 socket");
   chk(setSocketNonBlocking(fdIcmp4), "Failed to set non blocking on icmp4");
   patd_Icmp4.fdIcmp = fdIcmp4;
   epEvent.data.ptr = &patd_Icmp4;
   epEvent.events = EPOLLIN /* | EPOLLET */;
   chk(epoll_ctl(patd_Pool.fdPool, EPOLL_CTL_ADD, fdIcmp4, &epEvent),
       "Failed to add fdIcmp4 to EPoll");

   chk(fdIcmp6, "Failed to get a raw IPv6 socket");
   chk(setSocketNonBlocking(fdIcmp6), "Failed to set non blocking on icmp6");
//   chk(setsockopt(fdIcmp6, SOL_IPV6, IPV6_RECVERR, &iVal, sizeof(iVal)),
//       "Failed to set IPV6_RECVERR on icmp6");
   chk(setsockopt(fdIcmp6, IPPROTO_IPV6, IPV6_RECVPKTINFO, &iVal, sizeof(iVal)),
       "Failed to set IPV6_RECVPKTINFO on icmp6");
   patd_Icmp6.fdIcmp = fdIcmp6;
   epEvent.data.ptr = &patd_Icmp6;
   epEvent.events = EPOLLIN /* | EPOLLET */;
   chk(epoll_ctl(patd_Pool.fdPool, EPOLL_CTL_ADD, fdIcmp6, &epEvent),
       "Failed to add fdIcmp6 to EPoll");
   }

/*************************************************************************************************/
/** Deletes a PatdClient.
**************************************************************************************************/
void deletePatdClient (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "deletePatdClient: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus == STATUS_UNUSED) {
      return;
      }

   int iInx = 0;
   while ((iInx < MAX_CLIENTS) && (&apatd_Clients[ai_ClientList[iInx]] != pd)) {
      iInx += 1;
      }
   chk((MAX_CLIENTS - iInx - 1), "deletePatdClient: internal error: iInx == MAX_CLIENTS");

   i_ClientCount -= 1;
   int iTmp = ai_ClientList[i_ClientCount];              // swap ai_ClientList[i_ClientCount]
   ai_ClientList[i_ClientCount] = ai_ClientList[iInx];   // and  ai_ClientList[iInx]
   ai_ClientList[iInx] = iTmp;

   close(pd->fdClient);
   pd->iStatus = STATUS_UNUSED;
   pd->fdClient = -1;

   syslog(LOG_DEBUG, "deletePatdClient: DELETED: id=%i", pd->iId);
   }

/*************************************************************************************************/
/** Reads a Client Command.
**************************************************************************************************/
static void readClientCommand (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "readClientCommand: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   if (pd->iStatus != STATUS_NEW) {
      syslog(LOG_ERR, "readClientCommand: not STATUS_NEW");
      return;
      }

   ssize_t iSize = recv(pd->fdClient, aui8_Buffer, sizeof(aui8_Buffer), MSG_PEEK);
   if (iSize < 1) {
      syslog(LOG_DEBUG, "readClientCommand: no enought data size=%zi", iSize);
      return;
      }

   if (readAll(pd->fdClient, aui8_Buffer, 1)) {
      syslog(LOG_ERR, "readClientCommand: error read socket=%i id=%i", pd->fdClient, pd->iId);
      deletePatdClient(pd);
      return;
      }

   switch (aui8_Buffer[0]) {
      case COMMAND_IP4:
         pd->iStatus = STATUS_READ_IP4;
         clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);
         break;

      case COMMAND_IP6:
         pd->iStatus = STATUS_READ_IP6;
         clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);
         break;

      default:
         syslog(LOG_ERR, "readClientCommand: invalid command %i", aui8_Buffer[0]);
         deletePatdClient(pd);
      }

   rearmClient(pd);
   }

/*************************************************************************************************/
/** Allocates a new PatdClient.
**************************************************************************************************/
static int newPatdClient (int fdClient)
   {
   syslog(LOG_DEBUG, "newPatdClient: socket=%i", fdClient);

   chk(setSocketNonBlocking(fdClient), "Failed to set non blocking on client");
   if (i_ClientCount >= MAX_CLIENTS) {
      chk(-1, "Internal error: i_ClientCount >= MAX_CLIENTS");
      }

   int iNew = i_ClientCount + (random() % (MAX_CLIENTS - i_ClientCount));

   int iTmp = ai_ClientList[i_ClientCount];              // swap ai_ClientList[i_ClientCount]
   ai_ClientList[i_ClientCount] = ai_ClientList[iNew];   // and  ai_ClientList[iNew]
   ai_ClientList[iNew] = iTmp;

   PatdClient * pd = &apatd_Clients[ai_ClientList[i_ClientCount]];
   i_ClientCount += 1;

   pd->fdClient = fdClient;
   clock_gettime(CLOCK_MONOTONIC, &pd->tsLast);
   pd->iStatus = STATUS_NEW;

   struct epoll_event epEvent;
   epEvent.data.ptr = pd;
   epEvent.events = EPOLLIN | EPOLLET;

   if (epoll_ctl(patd_Pool.fdPool, EPOLL_CTL_ADD, fdClient, &epEvent)) {
      syslog(LOG_WARNING, "newPatdClient: cannot add to epool: socket=%i id=%i",
                          pd->fdClient, pd->iId);
      deletePatdClient(pd);
      return -1;
      }

   syslog(LOG_DEBUG, "newPatdClient: NEW socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   return 0;
   }

/*************************************************************************************************/
/** processServer.
**************************************************************************************************/
static void processServer (void)
   {
   syslog(LOG_DEBUG, "processServer: start");
   while (1) {
      if (i_ClientCount >= MAX_CLIENTS) {
         syslog(LOG_WARNING, "processServer: max clients");
         return;
         }

      struct sockaddr saClient;
      socklen_t iSize = sizeof(saClient);
      int fdClient = accept(patd_Server.fdServer, &saClient, &iSize);

      if (fdClient < 0) {
         syslog(LOG_DEBUG, "processServer: no more");
         return;
         }

      if (newPatdClient(fdClient)) {
         syslog(LOG_DEBUG, "processServer: error?");
         return;
         }
      }
   }

/*************************************************************************************************/
/** processClient.
**************************************************************************************************/
static void processClient (PatdClient * pd)
   {
   syslog(LOG_DEBUG, "processClient: socket=%i id=%i status=%i",
                     pd->fdClient, pd->iId, pd->iStatus);
   ssize_t iSize = recv(pd->fdClient, aui8_Buffer, sizeof(aui8_Buffer), MSG_PEEK);

   if (iSize <= 0) {                 // No more Data
      if ((errno == EAGAIN) || (errno == EWOULDBLOCK)) {        // ok
         return;
         }
      syslog(LOG_DEBUG, "processClient: close socket=%i id=%i size %zi: %i %s",
                        pd->fdClient, pd->iId, iSize, errno, strerror(errno));
      deletePatdClient(pd);
      return;
      }

   switch (pd->iStatus) {
      case STATUS_NEW:           readClientCommand(pd);     break;
      case STATUS_READ_IP4:      ip4ReadConfig(pd);         break;
      case STATUS_READ_IP6:      ip6ReadConfig(pd);         break;
      default:
         syslog(LOG_DEBUG, "processClient: strange status socket=%i id=%i status=%i",
                           pd->fdClient, pd->iId, pd->iStatus);
         deletePatdClient(pd);
      }
   }

/*************************************************************************************************/
/** processTimeOut.
**************************************************************************************************/
static void processTimeOut (void)
   {
   struct timespec tsNow;
   clock_gettime(CLOCK_MONOTONIC, &tsNow);

   struct timespec tsTime;
   for (int i = 0; i < i_ClientCount; i += 1) {
      PatdClient * pd = &apatd_Clients[ai_ClientList[i]];
      tsDelta(&tsNow, &pd->tsLast, &tsTime);
      if ((pd->iStatus == STATUS_IP4) && (tsCompare(&tsTime, &pd->tsTimeOut) > 0)) {
         ip4TimeOut(pd);
         }
      if ((pd->iStatus == STATUS_IP6) && (tsCompare(&tsTime, &pd->tsTimeOut) > 0)) {
         ip6TimeOut(pd);
         }
      if (/*(pd->iStatus == STATUS_NEW) && */ tsCompare(&tsTime, &ts_ClientTimeOut) > 0) {
         deletePatdClient(pd);
         }
      }
   }

/*************************************************************************************************/
/** Parses the options.
**************************************************************************************************/
static void parseOptions (int argc, const char *argv[])
   {
   for (int i = 1; i < argc; i += 1) {
      if (!strcmp(argv[i], "--address")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "%s: missing server argument\n", argv[0]);
            exit(EXIT_FAILURE);
            }
         str_Host = argv[i];
         }
      else if (!strcmp(argv[i], "--port")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "%s: missing port argument\n", argv[0]);
            exit(EXIT_FAILURE);
            }
         str_Port = argv[i];
         }
      else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
         fprintf(stderr,
            "Usage: %s [option]... [host]...\n"
            "Options:\n"
            "        --address ADDR    set the address to listen to (default: localhost)\n"
            "        --port PORT       set the port to listen to (default: 7865)\n"
            "   -f,  --foreground      run in foreground\n"
            "   -p,  --pidfile         write the pidfile\n"
            "   -h,  --help            display this help and exit\n"
            "   -q,  --quiet           log less messages to syslog\n"
            "   -v,  --verbose         log more messages to syslog\n"
            "   -V,  --version         display version information and exit\n",
            argv[0]);
         exit(EXIT_FAILURE);
         }
      else if (!strcmp(argv[i], "-f") || !strcmp(argv[i], "--foreground")) {
         i_Foreground = 1;
         }
      else if (!strcmp(argv[i], "-p") || !strcmp(argv[i], "--pidfile")) {
         i += 1;
         if (i == argc) {
            fprintf(stderr, "%s: missing pidfile argument\n", argv[0]);
            exit(EXIT_FAILURE);
            }
         str_Pid = argv[i];
         }
      else if (!strcmp(argv[i], "-q") || !strcmp(argv[i], "--quiet")) {
         i_LogLevel = (i_LogLevel > LOG_ERR) ? i_LogLevel - 1 : i_LogLevel;
         }
      else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--verbose")) {
         i_LogLevel = (i_LogLevel < LOG_DEBUG) ? i_LogLevel + 1 : i_LogLevel;
         }
      else if (!strcmp(argv[i], "-V") || !strcmp(argv[i], "--version")) {
         fprintf(stderr,
            "%s - MasarLabs PingAndTraceDaemon " VERSION "\n"
            "\n"
            "  Copyright (C) 2013-2014 MasarLabs.com\n"
            "  Written by Maurizio Sartori <masar@MasarLabs.com>\n"
            "\n"
            "  License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n"
            "  This is free software: you are free to change and redistribute it.\n"
            "  There is NO WARRANTY, to the extent permitted by law.\n",
            argv[0]);
         exit(EXIT_FAILURE);
         }
      else {
         fprintf(stderr, "%s: invalid parameter: %s\n", argv[0], argv[i]);
         exit(EXIT_FAILURE);
         }
      }
   }

/*************************************************************************************************/
/** Main entry point.
**************************************************************************************************/
int main (int argc, const char *argv[])
   {
   parseOptions(argc, argv);

   setlogmask(LOG_UPTO(i_LogLevel));
//   openlog(argv[0], LOG_CONS | LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_DAEMON);
   openlog(argv[0], LOG_PID | LOG_NDELAY, LOG_DAEMON);
   syslog(LOG_NOTICE, "Program started by User %d with loglevel %i", getuid(), i_LogLevel);
//   syslog(LOG_ERR,      "LOG_ERR     is enabled.");
//   syslog(LOG_WARNING,  "LOG_WARNING is enabled.");
//   syslog(LOG_NOTICE,   "LOG_NOTICE  is enabled.");
//   syslog(LOG_INFO,     "LOG_INFO    is enabled.");
//   syslog(LOG_DEBUG,    "LOG_DEBUG   is enabled.");

   chk(setSignal(SIGINT, signalHandlerIncrement), "Failed to set signal handler");

   initializeData();

   setupPatdPool();
   setupPatdIcmp();
   setupPatdServer(str_Host, str_Port);

   if (!i_Foreground) {
      chk(daemon(0,0), "Failed to daemonize");
      }
   syslog(LOG_NOTICE, "New pid %ld, old was %ld", (long) getpid(), (long) ui32_Pid);
   ui32_Pid = getpid();

   if (str_Pid != NULL) {
      chk(writePidfile(str_Pid), "Failed to write pidfile");
      }

   while (sig_SignalFlag == 0) {
      errno = 0;
      int iCount = epoll_wait(patd_Pool.fdPool, aep_Events, POOL_EVENTS, (i_ClientCount == 0) ? -1 : 100);
      if (sig_SignalFlag) {
         syslog(LOG_INFO, "Break: sig_SignalFlag=%i", sig_SignalFlag);
         break;
         }
      if ((iCount < 0) && (errno != EINTR)) {
         syslog(LOG_ERR, "Failed to wait epoll %i: %s", errno, strerror(errno));
         break;
         }

      for (int i = 0; i < iCount; i += 1) {
         struct epoll_event * pepEvent = &aep_Events[i];
         if ((pepEvent->events & (EPOLLERR | EPOLLHUP)) || (!(pepEvent->events & EPOLLIN))) {
            syslog(LOG_ERR, "Error event epoll %x %p", pepEvent->events, pepEvent->data.ptr);
            int       error = 0;
            socklen_t errlen = sizeof(error);
            if (getsockopt(patd_Icmp6.fdIcmp, SOL_SOCKET, SO_ERROR, (void *)&error, &errlen) == 0)
               {
               syslog(LOG_ERR, "error = %s\n", strerror(error));
               }
            }
         else if (pepEvent->data.ptr == &patd_Icmp4) {
            ip4Process();
            }
         else if (pepEvent->data.ptr == &patd_Icmp6) {
            ip6Process();
            }
         else if (pepEvent->data.ptr == &patd_Server) {
            processServer();
            }
         else {
            processClient(pepEvent->data.ptr);
            }
         }
      processTimeOut();
      }

   close(patd_Icmp4.fdIcmp);
   close(patd_Icmp6.fdIcmp);
   close(patd_Server.fdServer);

   syslog(LOG_NOTICE, "Program exit: %i", EXIT_SUCCESS);
   closelog();
   return EXIT_SUCCESS;
   }

/*************************************************************************************************/
/*************************************************************************************************/
